import 'package:flutter/material.dart';
import 'package:mobile_ui/src/controls/spacing.dart';

import '../mobile_ui_container.dart';

typedef ValueValidator<T> = String? Function(
    T value, String? errorCode, String? errorMsg);
enum ControlPosition { leading, trailing }
enum FieldOrientation { vertical, horizontal }

class FormInputLabelStyle {
  const FormInputLabelStyle({this.color, this.maxLine});

  final Color? color;
  final int? maxLine;
}

class FormInputBackgroundStyle {
  const FormInputBackgroundStyle(
      {this.backgroudColor, this.borderColor, this.borderWidth});

  final Color? backgroudColor;
  final Color? borderColor;
  final double? borderWidth;
}

const FormInputBackgroundStyle defalutFormInputBackgroundStyle =
    FormInputBackgroundStyle(backgroudColor: Colors.white);

const FormInputLabelStyle defalutFormInputLabelStyle =
    FormInputLabelStyle(color: Colors.black);

class FormInputBase<T> extends FormField<T> {
  final bool? isRequired;
  final String? labelText;
  final String? hintText;
  final EdgeInsetsGeometry contentPadding;
  final EdgeInsetsGeometry margin;
  final double spacingFlactor;
  final bool enabled;
  final bool readOnly;
  final TextAlign labelAlign;
  final FieldOrientation orientation;
  final ControlPosition controlPosition;
  final ValueNotifier<T?>? valueController;
  final ValueChanged<T?>? onChanged;
  final Icon? prefixIcon;
  final bool? isDense;
  final int errorMaxLines;
  final double? labelWidth;
  final FormInputLabelStyle? labelStyle;

  FormInputBase(
      {Key? key,
      required FormFieldBuilder<T> builder,
      ValueValidator<T?>? validator,
      T? initialValue,
      this.valueController,
      this.prefixIcon,
      this.isRequired,
      this.labelText,
      this.hintText,
      this.contentPadding = const EdgeInsets.all(8),
      this.margin = const EdgeInsets.symmetric(vertical: 8),
      this.spacingFlactor = 1,
      required bool enabled,
      this.isDense,
      bool? readOnly,
      this.labelAlign = TextAlign.left,
      this.orientation = FieldOrientation.vertical,
      //Only valid when the direction is equal to horizontal
      this.controlPosition = ControlPosition.trailing,
      AutovalidateMode? autovalidateMode,
      FormFieldSetter<T>? onSaved,
      this.onChanged,
      this.errorMaxLines = 1,
      this.labelWidth,
      this.labelStyle})
      : assert(initialValue == null || valueController == null),
        enabled = enabled,
        readOnly = (enabled == false || readOnly == true),
        super(
            builder: builder,
            initialValue:
                valueController != null ? valueController.value : initialValue,
            onSaved: onSaved,
            enabled: enabled,
            autovalidateMode: autovalidateMode,
            validator: (value) {
              String? errorCode;
              String? errorMsg;
              final isNullOrEmpty = (value == null ||
                  (value is String && value.isEmpty) ||
                  (value is List && value.isEmpty));
              if (isNullOrEmpty) {
                if (isRequired == true) {
                  errorCode = 'required';
                  errorMsg = '${labelText ?? "This field"} is required.';
                }
              }
              if (validator != null)
                return validator(value, errorCode, errorMsg);
              return errorMsg;
            });

  Widget buildField(FormFieldState<T> field, Widget child,
      {FormInputBackgroundStyle? backgroudStyle}) {
    final theme = Theme.of(field.context);
    final uiConfig = MobileUIContainer.of(field.context);
    final bgStyle = backgroudStyle ??
        uiConfig?.formInputBackgroundStyle ??
        defalutFormInputBackgroundStyle;
    final formLabelStyle = labelStyle ??
        uiConfig?.formInputLabelStyle ??
        defalutFormInputLabelStyle;
    if (orientation == FieldOrientation.vertical) {
      return buildVField(
          theme,
          buildDecorator(
            field,
            child,
            backgroundStyle: bgStyle,
          ),
          formLabelStyle);
    } else {
      return buildHField(
          theme,
          buildDecorator(
            field,
            child,
            backgroundStyle: bgStyle,
          ),
          formLabelStyle);
    }
  }

  Widget buildHField(
      ThemeData theme, Widget child, FormInputLabelStyle formLabelStyle) {
    final hasLabel = (labelText != null && labelText!.isNotEmpty);
    var children = <Widget>[];
    if (hasLabel) {
      if (controlPosition == ControlPosition.leading) {
        children.addAll([
          Expanded(child: child),
          Spacing(spacingFlactor),
          buildLabelText(theme, formLabelStyle)
        ]);
      } else {
        children.addAll([
          buildLabelText(theme, formLabelStyle),
          Spacing(spacingFlactor),
          Expanded(child: child)
        ]);
      }
    }
    return Padding(
        padding: margin,
        child: hasLabel
            ? Row(
                children: children,
              )
            : child);
  }

  Widget buildVField(
      ThemeData theme, Widget child, FormInputLabelStyle formLabelStyle) {
    final hasLabel = (labelText != null && labelText!.isNotEmpty);
    var children = <Widget>[];
    if (hasLabel) {
      children.addAll([
        buildLabelText(theme, formLabelStyle),
        Spacing(spacingFlactor),
      ]);
    }
    return Padding(
        padding: margin,
        child: hasLabel
            ? Column(
                crossAxisAlignment: CrossAxisAlignment.stretch,
                children: [...children, child])
            : child);
  }

  Widget buildLabelText(ThemeData theme, FormInputLabelStyle formLabelStyle) {
    late Widget labelWidget;
    final labelColor = formLabelStyle.color;
    final text = Text(
      labelText ?? '',
      textAlign: labelAlign,
      maxLines: formLabelStyle.maxLine,
      style: enabled
          ? theme.textTheme.subtitle1!.copyWith(color: labelColor)
          : theme.textTheme.subtitle1!.copyWith(color: theme.disabledColor),
    );
    if (isRequired == true) {
      var mainAxisAlignment = MainAxisAlignment.start;
      if (labelAlign == TextAlign.center) {
        mainAxisAlignment = MainAxisAlignment.center;
      } else if (labelAlign == TextAlign.right) {
        mainAxisAlignment = MainAxisAlignment.end;
      }
      labelWidget = Row(
        mainAxisAlignment: mainAxisAlignment,
        mainAxisSize: MainAxisSize.min,
        children: [
          Text(
            '* ',
            textAlign: labelAlign,
            style: enabled
                ? theme.textTheme.subtitle1!.copyWith(color: Colors.red)
                : theme.textTheme.subtitle1!
                    .copyWith(color: theme.disabledColor),
          ),
          text,
        ],
      );
    } else {
      labelWidget = text;
    }
    if (labelWidth != null) {
      return SizedBox(width: labelWidth, child: labelWidget);
    }
    return labelWidget;
  }

  Widget buildText(ThemeData theme, String? text,
      {Color? color, bool? softWrap = true}) {
    return Text(
      text ?? '',
      textAlign: TextAlign.left,
      softWrap: softWrap,
      style: enabled
          ? (color == null
              ? theme.textTheme.subtitle1
              : theme.textTheme.subtitle1!.copyWith(color: color))
          : theme.textTheme.subtitle1!.copyWith(color: theme.disabledColor),
    );
  }

  Widget buildIcon(ThemeData theme, IconData icon,
      {GestureTapCallback? onTap, Color? enabledColor}) {
    final color = (enabled && !readOnly)
        ? enabledColor ?? theme.primaryColor
        : theme.disabledColor;
    final iconWidget = Icon(
      icon,
      color: color,
    );
    if (onTap != null && !readOnly) {
      return InkWell(
        onTap: onTap,
        child: iconWidget,
      );
    }
    return iconWidget;
  }

  bool getIsNullOrEmpty(T? value) {
    return (value == null ||
        (value is String && value.isEmpty) ||
        (value is List && value.isEmpty));
  }

  Widget buildDecorator(FormFieldState<T> field, Widget child,
      {FormInputBackgroundStyle? backgroundStyle}) {
    final effectiveDecoration =
        InputDecoration(contentPadding: EdgeInsets.all(0)).copyWith(
            hintText: hintText,
            border: OutlineInputBorder(borderSide: BorderSide.none));
    final isNullOrEmpty = getIsNullOrEmpty(field.value);

    if (backgroundStyle == null) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: child);
    }
    final hasBorder = backgroundStyle.borderWidth != null;
    final theme = Theme.of(field.context);
    if (hasBorder) {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              fillColor: backgroundStyle.backgroudColor,
              filled: backgroundStyle.backgroudColor != null,
              prefixIcon: prefixIcon,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense,
              contentPadding: contentPadding,
              hintText: hintText,
              errorBorder: OutlineInputBorder(
                  borderSide: BorderSide(
                      color: theme.errorColor,
                      width: backgroundStyle.borderWidth ?? 1)),
              enabledBorder: OutlineInputBorder(
                  borderSide: BorderSide(
                      color: backgroundStyle.borderColor ?? Color(0xFFEEEEEE),
                      width: backgroundStyle.borderWidth ?? 1))),
          isEmpty: isNullOrEmpty,
          child: child);
    } else {
      return InputDecorator(
          decoration: effectiveDecoration.copyWith(
              hintText: hintText,
              errorText: field.errorText,
              errorMaxLines: errorMaxLines,
              isDense: isDense),
          isEmpty: isNullOrEmpty,
          child: Container(
              constraints: BoxConstraints(minHeight: 48),
              color: backgroundStyle.backgroudColor,
              padding: contentPadding,
              child: child));
    }
  }

  @override
  FormFieldState<T> createState() => FormInputStateBase<T>();
}

class FormInputStateBase<T> extends FormFieldState<T> {
  ValueNotifier<T?>? _valueController;

  ValueNotifier<T?>? get effectiveValueController =>
      widget.valueController ?? _valueController;

  @override
  FormInputBase<T> get widget => super.widget as FormInputBase<T>;

  @override
  void initState() {
    super.initState();
    if (widget.valueController == null) {
      _valueController = ValueNotifier<T?>(widget.initialValue);
    } else {
      widget.valueController!.addListener(handleControllerChanged);
    }
  }

  @override
  bool validate() {
    saveValue();
    return super.validate();
  }

  @override
  void didUpdateWidget(FormInputBase<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.valueController != oldWidget.valueController) {
      oldWidget.valueController?.removeListener(handleControllerChanged);
      widget.valueController?.addListener(handleControllerChanged);

      if (oldWidget.valueController != null && widget.valueController == null) {
        _valueController = ValueNotifier<T?>(oldWidget.valueController!.value);
      }
      if (widget.valueController != null) {
        setValue(widget.valueController!.value);
        if (oldWidget.valueController == null) _valueController = null;
      }
    }
  }

  @override
  void save() {
    saveValue();
    super.save();
  }

  @override
  void dispose() {
    widget.valueController?.removeListener(handleControllerChanged);
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    setState(() {
      effectiveValueController!.value = widget.initialValue;
    });
  }

  @protected
  void handleControllerChanged() {
    changeValue(effectiveValueController!.value);
  }

  @protected
  void saveValue() {
    // changeValue(effectiveValueController!.value);
  }

  void changeValue(T? v) {
    if (effectiveValueController!.value != v) {
      effectiveValueController!.value = v;
    }
    if (v != value) {
      didChange(v);
      if (widget.onChanged != null) {
        widget.onChanged!(v);
      }
    }
  }
}

class ValueController<T> extends ValueNotifier<T> {
  ValueController(T value) : super(value);

  ValueController.fromValue(T value) : super(value);
}
